"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Defines the structure of protobuf messages for setting camera control status
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions
DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _EnumCameraControlStatus:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _EnumCameraControlStatusEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_EnumCameraControlStatus.ValueType],
    builtins.type,
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    CAMERA_IDLE: _EnumCameraControlStatus.ValueType
    CAMERA_CONTROL: _EnumCameraControlStatus.ValueType
    "Can only be set by camera, not by third-party"
    CAMERA_EXTERNAL_CONTROL: _EnumCameraControlStatus.ValueType

class EnumCameraControlStatus(_EnumCameraControlStatus, metaclass=_EnumCameraControlStatusEnumTypeWrapper): ...

CAMERA_IDLE: EnumCameraControlStatus.ValueType
CAMERA_CONTROL: EnumCameraControlStatus.ValueType
"Can only be set by camera, not by third-party"
CAMERA_EXTERNAL_CONTROL: EnumCameraControlStatus.ValueType
global___EnumCameraControlStatus = EnumCameraControlStatus

class RequestSetCameraControlStatus(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    CAMERA_CONTROL_STATUS_FIELD_NUMBER: builtins.int
    camera_control_status: global___EnumCameraControlStatus.ValueType
    "Declare who is taking control of the camera"

    def __init__(self, *, camera_control_status: global___EnumCameraControlStatus.ValueType | None = ...) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["camera_control_status", b"camera_control_status"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["camera_control_status", b"camera_control_status"]
    ) -> None: ...

global___RequestSetCameraControlStatus = RequestSetCameraControlStatus
